#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>


/*
** notes:
**
** The code is really too long for one file, but I have at least seperated it
** into some logical sections.
**
** The basic functionality is to read in a list of filenames containing
** SIPD strategies and to play them against each other in a tournament.
**
** Extra features include:
** 
** - Random initial formulae.
** - Proportional survival tournament (aka ecological tournament).
**   - takes strategies and initial quantities to begin with
**   - plays many rounds of SIPD, adjusting the proportions after each.
** - Evolutionary tournament
**   - plays many round of SIPD, updating the strategies after each.
*/


#define DEBUG  1

#define AND          1 /* boring constants */
#define OR           2
#define NOT          3
#define ALL_AGENTS   4
#define EXISTS_AGENT 5
#define ALL_TIMES    6
#define EXISTS_TIME  7
#define ATOM         8
#define RANDOM       9

#define FALSE  0
#define TRUE   1

#define LEFT   0
#define RIGHT  1

#define COOPERATE 0
#define DEFECT    1

#define ROW    -2
#define COLUMN -3

#define PREV   -4

#define NONE  -1

#define SUCKER     0 /* payoff matrix - change if desired */
#define PUNISHMENT 1
#define REWARD     3
#define TEMPTATION 5

#define MAX_FILE_LENGTH      1000
#define MAX_FILE_NAME_LENGTH  200
#define MAX_INT_LENGTH          7
#define MAX_POP               100
#define MAX_PROP_POP         1000 /* max pop for proportional games */
#define GAME_LENGTH           100 /* # of turns in an SIPD game */
#define MAX_TIME              101 /* XXX I have no idea why this is one higher */
#define MAX_ROUNDS            100 /* # of rounds of SIPD for ecological / genetic */

#define STRATEGY_LOCATION "strategies/" /* path of dir containing strats */

#define ROUND_SIZE       (MAX_POP * MAX_POP)
#define HISTORY_SIZE     (ROUND_SIZE * MAX_TIME)

#define POP_FOR_RANDOM_FORMULAE 10

#define MAX_AGENT_VARS    10 /* "defghijklm"         */
#define MAX_AGENT_SYMBOLS 12 /* "cr" ++ "defghijklm" */
#define MAX_TIME_VARS      8 /* "stuvwxyz"           */
#define MAX_TIME_SYMBOLS   9 /* "p"  ++ "stuvwxyz"   */

#define INVERSE_MUT_RATE  10

#define SIMPLIFY_FORMULAE 1 /* whether or not to remove useless double neg
                               and unused quantifiers from formulae        */

#define ALLVARS 1 /* lets the variables take on all values (of their type)
                     including the constants r, c, t.                      
                     It is stuck on at the moment.                          */

#define AND_CHANCE       4
#define OR_CHANCE        4
#define NOT_CHANCE       4
#define ALL_A_CHANCE     4
#define EXISTS_A_CHANCE  4
#define ALL_T_CHANCE     4
#define EXISTS_T_CHANCE  4
#define ATOM_CHANCE      8

#define MAX_DEPTH              5 /* limits for random formula generation */
#define MAX_QUANTIFICATION 10000

typedef struct Formula_s Formula;
struct Formula_s {
  int       type;
  int       variable;    /* for quantifiers */
  int       player1;     /* If a variable (ie. player1, player1 or time) */
  int       player2;     /* is negative, it represents a relative player */
  int       time;        /* (ie. ROW, COLUMN, PREV). Otherwise it is the */
  int       time_mod;    /* name of a variable (eg 't', 'x')             */
  double    chance;
  Formula * left;
  Formula * right;
};

typedef struct Result_s Result;
struct Result_s {
  int t;
  int r; 
  int p;
  int s;
  int score;
};

typedef struct Formula_List_s Formula_List;
struct Formula_List_s {
  Formula *    formula;
  Formula_List * next;
};

typedef int agent;

typedef char * string;


/* simulation functions */

void      simulate(int *, Formula **, int);
void      simulate_prop(int *, Formula **, int *, int);
int       evaluate(int *, int *, int *, Formula *, int, agent, agent, int);
int       evaluate_prop(int *, int *, int *, Formula *,
                        int *, int, agent, agent, int);
void      add_action(int *, agent, agent, int, int);
void      add_action_prop(int *, agent, agent, int, int);
int       get_action(int *, agent, agent, int);
int       get_action_prop(int *, agent, agent, int);
void      get_Results(int *, Result *, int);
void      get_Results_prop(int *, Result *, int *, int);

void      sub_agent(int *, int, int);
void      unsub_agent(int *, int);
int       get_agent_sub(int *, int);
void      sub_time(int *, int, int);
void      unsub_time(int *, int);
int       get_time_sub(int *, int);

void      update_proportions(Result *, int *, int);

/* formula functions */

Formula * new_Formula(int);
void      free_Formula(Formula *);

void      init_random_Formulae(Formula **, int);
Formula * random_Formula(string, string, int, int);
agent     random_agent(string, agent);
int       random_time(string);

void      init_variables(string, string);
void      add_variable(string, int);

void      fix_Formulae(Formula **, int, int);
void      fix_Formula(Formula *, string, string);
void      replace_var(Formula *, int, int);
int       useless_var(Formula *, int);
void      simplify_Formula(Formula **);
void      simplify_Formula_2(Formula *, Formula *, int);


int       create_new_agent_var(string);
int       create_new_time_var(string);



/* genetic functions */

void           update_Formulae(Result *, Formula **, int);

Formula *      cross_over(Formula *, Formula *);
agent          choose_parent(Result *, int);
Formula_List * flatten_Formula(Formula *, Formula_List *);
Formula_List * cons_Formula(Formula *, Formula_List *);
int            length_Formula_List(Formula_List *);
Formula *      choose_random_Formula(Formula *);
Formula *      copy_Formula(Formula *);

void           mutate_Formula(Formula *);
void           mutate_an_atom(Formula *, string, string);
void           mutate_atom(Formula *, string, string);

/* input functions */

int       init_Formulae(Formula **);
int       init_Formulae_prop(Formula **, int *);
Formula * read_Formula(const char *);
Formula * make_Formula(char *, int, int);
int       make_time_mod(char *, int, int);
double    get_chance(char *, int, int);
int       find_conjunct(char *, int, int);

/* printing functions */

void      print_Formula(Formula *);
void      print_Formulae(Formula **, int);
void      print_quantities(int *, int);
void      print_Results(Result *, int);
void      print_Results_prop(Result *, int *, int);
void      print_actions(int *, int, agent);
void      print_Formula_List(Formula_List *);

/* utility functions */

void      error(const char *);
void      debug(const char *);


int
main(int argc, char *argv[])
{
  int history[HISTORY_SIZE];
  Result results[MAX_POP];
  
  int pop;
  int player;
	Formula * formulae[MAX_POP];
  int quantities[MAX_POP];
  int round;
  int evolve = FALSE;
  int random_Formulae = FALSE;
  int proportional = FALSE;

  srand(42);

  /* random seed options */
  if(argc >= 3 && strcmp(argv[1],"-s")==0)
    srand(atoi(argv[2]));
  if(argc >= 4 && strcmp(argv[2],"-s")==0)
    srand(atoi(argv[3]));
  if(argc >= 5 && strcmp(argv[3],"-s")==0)
    srand(atoi(argv[4]));

  /* evolution option */
  if(argc >= 2 && strcmp(argv[1],"-e")==0)
    evolve = TRUE;
  if(argc >= 3 && strcmp(argv[2],"-e")==0)
    evolve = TRUE;
  if(argc >= 4 && strcmp(argv[3],"-e")==0)
    evolve = TRUE;
  if(argc >= 5 && strcmp(argv[4],"-e")==0)
    evolve = TRUE;

  /* random formulae option */
  if(argc >= 2 && strcmp(argv[1],"-r")==0)
    random_Formulae = TRUE;
  if(argc >= 3 && strcmp(argv[2],"-r")==0)
    random_Formulae = TRUE;
  if(argc >= 4 && strcmp(argv[3],"-r")==0)
    random_Formulae = TRUE;
  if(argc >= 5 && strcmp(argv[4],"-r")==0)
    random_Formulae = TRUE;

  /* proportial representation option */
  if(argc >= 2 && strcmp(argv[1],"-p")==0)
    proportional = TRUE;
  if(argc >= 3 && strcmp(argv[2],"-p")==0)
    proportional = TRUE;
  if(argc >= 4 && strcmp(argv[3],"-p")==0)
    proportional = TRUE;
  if(argc >= 5 && strcmp(argv[4],"-p")==0)
    proportional = TRUE;
  
  if (random_Formulae){
    pop = POP_FOR_RANDOM_FORMULAE;
    init_random_Formulae(formulae, pop);    
  }
  else if (proportional){
    pop = init_Formulae_prop(formulae,quantities);
  }
  else    
    pop = init_Formulae(formulae);


  /* the tournament: length MAX_ROUNDS for evolution, 1 round without it */
  if(evolve)
    for(round = 0; round < MAX_ROUNDS; round++){
      /* if(round == 0 || round%10 == 9) */
      print_Formulae(formulae, pop);
      simulate(history, formulae, pop);
      get_Results(history, results, pop);
      /* if(round == 0 || round%10 == 9) */
      print_Results(results, pop);
      update_Formulae(results, formulae, pop);
    }
  else if(proportional){
    for(round = 0; round < MAX_ROUNDS; round++){
      simulate_prop(history, formulae, quantities, pop);
      get_Results_prop(history, results, quantities, pop);
      print_Results_prop(results, quantities, pop);
    /*  for(player = 0; player < pop; player++)
        fprintf(stderr,"%d\t",quantities[player]); */
      fprintf(stderr,"\n");
      update_proportions(results, quantities, pop);
    }
  } else{
    if(random_Formulae)
      print_Formulae(formulae, pop);
    simulate(history, formulae, pop);
    get_Results(history, results, pop);
    print_Results(results, pop);
  }

  return 0; 
}

/*************************************************************************

      simulation functions

*************************************************************************/

void
simulate(int * history, Formula ** formulae, int pop)
{
  int time;
  int row, column;
  int action;
  int i;
  int a_subs[MAX_AGENT_VARS];
  int t_subs[MAX_TIME_VARS];
  Formula * fixed_formulae[MAX_POP];

  for(i = 0; i < HISTORY_SIZE; i++)
    history[i] = NONE;

  /* copies all formulae and fixes them, then uses these for the eval */
  for(i = 0; i < MAX_POP; i++)
    fixed_formulae[i] = NULL;
  for(i = 0; i < pop; i++)
    fixed_formulae[i] = copy_Formula(formulae[i]);
  fix_Formulae(fixed_formulae,pop,TRUE);
  

  for(time = 0; time < GAME_LENGTH; time++){
    for(row = 0; row < pop; row++)
      for(column = 0; column < pop; column++)
        if(row!=column){
	  /*
          printf("<");
          print_Formula(formulae[row]);
          printf(">\n");
	  */
          for(i = 0; i < MAX_AGENT_VARS; i++)
            a_subs[i] = NONE;
          for(i = 0; i < MAX_TIME_VARS; i++)
            t_subs[i] = NONE;
          action = evaluate(history,a_subs,t_subs,fixed_formulae[row],pop,row,column,time-1);
          add_action(history,row,column,time,action);
        }
    }

  /* frees the fixed formulae */
  free_Formula(fixed_formulae[i]);

  return;
}

void
simulate_prop(int * history, Formula ** formulae, int * quantities, int pop)
{
  int time;
  int row, column;
  int action;
  int i;
  int a_subs[MAX_AGENT_VARS];
  int t_subs[MAX_TIME_VARS];
  Formula * fixed_formulae[MAX_POP];

  for(i = 0; i < HISTORY_SIZE; i++)
    history[i] = NONE;

  /* copies all formulae and fixes them, then uses these for the eval */
  for(i = 0; i < MAX_POP; i++)
    fixed_formulae[i] = NULL;
  for(i = 0; i < pop; i++)
    fixed_formulae[i] = copy_Formula(formulae[i]);
  fix_Formulae(fixed_formulae,pop,TRUE);

  for(time = 0; time < GAME_LENGTH; time++)
    for(row = 0; row < pop; row++)
      for(column = 0; column < pop; column++)
        if((row!=column && quantities[column] >= 1 && quantities[row] >= 1)
            || (row==column && quantities[column] >= 2)){
	  /*
          printf("<");
          print_Formula(formulae[row]);
          printf(">\n");
	  */
          for(i = 0; i < MAX_AGENT_VARS; i++)
            a_subs[i] = NONE;
          for(i = 0; i < MAX_TIME_VARS; i++)
            t_subs[i] = NONE;
          action = evaluate_prop(history,a_subs,t_subs,fixed_formulae[row],
                                 quantities,pop,row,column,time-1);
          add_action_prop(history,row,column,time,action);
        }

  /* frees the fixed formulae */
  free_Formula(fixed_formulae[i]);

  return;
}

/*
** evaluates whether a given formula is true
** (ie whether a given agent will defect for a given opponent,
**  time and history)
*/

int
evaluate(int * history, int * a_subs, int * t_subs, Formula * formula, int pop,
                               agent row, agent column, int prev_time)
{
  int result;
  agent player1 = ROW, player2 = ROW; /* gnuc */
  agent player;
  int time = 0;
  Formula * new_formula;

  switch(formula->type){
    case AND:
      result = evaluate(history, a_subs, t_subs, formula->left,
                        pop, row, column, prev_time)
              && evaluate(history, a_subs, t_subs, formula->right,
                          pop, row, column, prev_time);
      break;
    case OR:
      result = evaluate(history, a_subs, t_subs, formula->left,
                        pop, row, column, prev_time)
              || evaluate(history, a_subs, t_subs, formula->right,
                          pop, row, column, prev_time);
      break;
    case NOT:
      result = !evaluate(history, a_subs, t_subs, formula->left, pop, row, column, prev_time);
      break;
    case ALL_AGENTS:
      result = TRUE;
      for(player = 0; player < pop; player++)
        if(result==TRUE && (ALLVARS || (player!=row && player!=column))){
	  sub_agent(a_subs, formula->variable, player);
          result = evaluate(history, a_subs, t_subs, formula->left,
                            pop, row, column, prev_time);
          unsub_agent(a_subs,formula->variable);
	    /*
            new_formula = substitute_agent(formula, formula->variable, player);
            result = evaluate(history, a_subs, t_subs, new_formula,pop, row, column, prev_time);
            free_Formula(new_formula);
	    */
        }
      break;
    case EXISTS_AGENT:
      result = FALSE;
      for(player = 0; player < pop; player++)
        if(result==FALSE && (ALLVARS || (player!=row && player!=column))){
            sub_agent(a_subs, formula->variable, player);
            result = evaluate(history, a_subs, t_subs, formula->left,
                              pop, row, column, prev_time);
            unsub_agent(a_subs,formula->variable);
	    /*
            new_formula = substitute_agent(formula, formula->variable, player);
            result = evaluate(history, a_subs, t_subs, new_formula,pop, row, column, prev_time);
            free_Formula(new_formula);
	    */
        }
      break;
    case ALL_TIMES:
      result = TRUE;
      if(!ALLVARS && prev_time < 1){         /* should be removed soon */
        error("old code in evaluation: should not be in here");        
      }
      else
        for(time = 0; time < prev_time + ALLVARS; time++) /* hack*/
          if(result==TRUE){	   
            sub_time(t_subs,formula->variable,time);
            result = evaluate(history, a_subs, t_subs, formula->left,
                              pop, row, column, prev_time);
            unsub_time(t_subs,formula->variable);
	    /*
            new_formula = substitute_time(formula, formula->variable, time);
            result = evaluate(history, a_subs, t_subs, new_formula,pop, row, column, prev_time);
            free_Formula(new_formula);
            */
          }
      break;
    case EXISTS_TIME:
      result = FALSE;
      if(!ALLVARS && prev_time < 1){         /* should be removed soon */
        error("old code in evaluation: should not be in here");        
      }
      else
        for(time = 0; time < prev_time + ALLVARS; time++) /* hack */
          if(result==FALSE){
            sub_time(t_subs,formula->variable,time);
            result = evaluate(history, a_subs, t_subs, formula->left,
                              pop, row, column, prev_time);
            unsub_time(t_subs,formula->variable);
	    /*
            new_formula =  substitute_time(formula, formula->variable, time);
            result = evaluate(history, a_subs, t_subs, new_formula,pop, row, column, prev_time);
            free_Formula(new_formula);
	    */
          }
      break;
    case ATOM:
      if(formula->player1 == ROW)
        player1 = row;
      else if (formula->player1 == COLUMN)
        player1 = column;
      else
        player1 = get_agent_sub(a_subs,formula->player1);

      if(formula->player2 == ROW)
        player2 = row;
      else if (formula->player2 == COLUMN)
        player2 = column;
      else
        player2 = get_agent_sub(a_subs,formula->player2);
 
      if(formula->time == PREV)
        time = prev_time;
      else
        time = get_time_sub(t_subs,formula->time);

      time += formula->time_mod;
      result = get_action(history, player1, player2, time);
      break;
    case RANDOM:
      if(rand() < SHRT_MAX*formula->chance)
        result = TRUE;
      else
        result = FALSE;
      break;
    default:
      result = 0;
      fprintf(stderr,"%d\n",formula->type);
      error("Invalid contructor in formula (while evaluating)");
      
  }       

  return result;
}

/*
** evaluates whether a given formula is true
** for a proportional tournament
*/

int
evaluate_prop(int * history, int * a_subs, int * t_subs, Formula * formula,
              int * quantities, int pop, agent row, agent column, int prev_time)
{
  int result;
  agent player1 = ROW, player2 = ROW; /* gnuc */
  agent player;
  int time = 0;
  Formula * new_formula;

  switch(formula->type){
    case AND:
      result = evaluate_prop(history, a_subs, t_subs, formula->left,
                             quantities, pop, row, column, prev_time)
              && evaluate_prop(history, a_subs, t_subs, formula->right,
                               quantities, pop, row, column, prev_time);
      break;
    case OR:
      result = evaluate_prop(history, a_subs, t_subs, formula->left,
                             quantities, pop, row, column, prev_time)
              || evaluate_prop(history, a_subs, t_subs, formula->right,
                               quantities, pop, row, column, prev_time);
      break;
    case NOT:
      result = !evaluate_prop(history, a_subs, t_subs, formula->left,
                              quantities, pop, row, column, prev_time);
      break;
    case ALL_AGENTS:
      result = TRUE;
      for(player = 0; player < pop; player++)
        if(quantities[player] >= 1)
          if(result==TRUE && (ALLVARS || (player!=row && player!=column))){
            sub_agent(a_subs,formula->variable,player);
            result = evaluate_prop(history, a_subs, t_subs, formula->left,
                                   quantities, pop, row, column, prev_time);
            unsub_agent(a_subs,formula->variable);
            /*
            new_formula = substitute_agent(formula, formula->variable, player);
            result = evaluate_prop(history, new_formula, quantities, pop, row, column, prev_time);
            free_Formula(new_formula);
	    */
          }
      break;
    case EXISTS_AGENT:
      result = FALSE;
      for(player = 0; player < pop; player++)
        if(quantities[player] >= 1)
          if(result==FALSE && (ALLVARS || (player!=row && player!=column))){
            sub_agent(a_subs,formula->variable,player);
            result = evaluate_prop(history, a_subs, t_subs, formula->left,
                                   quantities, pop, row, column, prev_time);
            unsub_agent(a_subs,formula->variable);
	    /*
            new_formula = substitute_agent(formula, formula->variable, player);
            result = evaluate_prop(history, new_formula, quantities, pop, row, column, prev_time);
            free_Formula(new_formula);
	    */
          }
      break;
    case ALL_TIMES:
      result = TRUE;
      if(!ALLVARS && prev_time < 1){         /* should be removed soon */
        error("old code in evaluation: should not be in here");        
      }
      else
        for(time = 0; time < prev_time + ALLVARS; time++) /* hack */
          if(result==TRUE){
            sub_time(t_subs,formula->variable,time);
            result = evaluate_prop(history, a_subs, t_subs, formula->left,
                                   quantities, pop, row, column, prev_time);
            unsub_time(t_subs,formula->variable);
	    /*
            new_formula = substitute_time(formula, formula->variable, time);
            result = evaluate_prop(history, new_formula, quantities, pop, row, column, prev_time);
            free_Formula(new_formula);
	    */
          }
      break;
    case EXISTS_TIME:
      result = FALSE;
      if(!ALLVARS && prev_time < 1){         /* should be removed soon */
        error("old code in evaluation: should not be in here");        
      }
      else
        for(time = 0; time < prev_time + ALLVARS; time++) /* hack */
          if(result==FALSE){
            sub_time(t_subs,formula->variable,time);
            result = evaluate_prop(history, a_subs, t_subs, formula->left,
                                   quantities, pop, row, column, prev_time);
            unsub_time(t_subs,formula->variable);
	    /*
            new_formula =  substitute_time(formula, formula->variable, time);
            result = evaluate_prop(history, new_formula, quantities, pop, row, column, prev_time);
            free_Formula(new_formula);
	    */
          }
      break;
    case ATOM:
      if(formula->player1 == ROW)
        player1 = row;
      else if (formula->player1 == COLUMN)
        player1 = column;
      else
        player1 = get_agent_sub(a_subs,formula->player1);

      if(formula->player2 == ROW)
        player2 = row;
      else if (formula->player2 == COLUMN)
        player2 = column;
      else
        player2 = get_agent_sub(a_subs,formula->player2);
 
      if(formula->time == PREV)
        time = prev_time;
      else
        time = get_time_sub(t_subs,formula->time);

      time += formula->time_mod;
      result = get_action_prop(history, player1, player2, time);
      break;
    case RANDOM:
      if(rand() < SHRT_MAX*formula->chance)
        result = TRUE;
      else
        result = FALSE;
      break;
    default:
      result = 0;
      fprintf(stderr,"%d\n",formula->type);
      error("Invalid contructor in formula (while evaluating)");
      
  }       

  return result;
}

/*
** substitution functions:
**
** sub_time(t_subs, time_var, time)
** updates t_subs to instantiate time_var to time
**
** unsub_time(t_subs, time_var)
** updates t_subs to uninstantiate time_var
**
** get_time_sub(t_subs, time_var)
** gets the instantiation of time_var
*/

void 
sub_agent(int * a_subs, int player_var, int player)
{
  if(a_subs[player_var - 'd'] != NONE)
    debug("attempting to substitute two agents for the one variable");
  if(player_var - 'd' < 0 || player_var - 'd' >= MAX_AGENT_VARS)
    error("substituting an out of bounds agent variable");
  a_subs[player_var - 'd'] = player;
  return;
} 

void 
unsub_agent(int * a_subs, int player_var)
{
  if(a_subs[player_var - 'd'] == NONE)
    debug("attempting to unsubstitute a nonsubstitued agent");
  if(player_var - 'd' < 0 || player_var - 'd' >= MAX_AGENT_VARS)
    error("unsubstituting an out of bounds agent variable");
  a_subs[player_var - 'd'] = NONE;
  return;
} 

int 
get_agent_sub(int * a_subs, int player_var)
{
  if(player_var - 'd' < 0 || player_var - 'd' >= MAX_AGENT_VARS){
    fprintf(stderr,"getting a substitution for an out of bounds agent variable (%c)\n",player_var);
    error("eek!");
  }
  if(a_subs[player_var - 'd'] != NONE)
    return a_subs[player_var - 'd'];
  else
    error("tried to evaluate a raw variable agent");
}

void 
sub_time(int * t_subs, int time_var, int time)
{
  if(t_subs[time_var - 's'] != NONE)
    debug("attempting to substitute two times for the one variable");
  if(time_var - 's' < 0 || time_var - 's' >= MAX_TIME_VARS)
    error("substituting an out of bounds time variable");
  t_subs[time_var - 's'] = time;
  return;
} 

void 
unsub_time(int * t_subs, int time_var)
{
  if(t_subs[time_var - 's'] == NONE)
    debug("attempting to unsubstitute a nonsubstitued time");
  if(time_var - 's' < 0 || time_var - 's' >= MAX_TIME_VARS)
    error("unsubstituting an out of bounds time variable");
  t_subs[time_var - 's'] = NONE;
  return;
} 

int 
get_time_sub(int * t_subs, int time_var)
{
  if(time_var - 's' < 0 || time_var - 's' >= MAX_TIME_VARS)
    error("getting a substituion for an out of bounds time variable");
  if(t_subs[time_var - 's'] != NONE)
    return t_subs[time_var - 's'];
  else
    error("tried to evaluate a raw variable time");
}

/*
** adds info about what one player did to another at a given time
*/

void
add_action(int * history, agent player1, agent player2, int time, int action)
{
  if(time < 0)
    error("adding action at negative time");
  
  if(player1 == player2)
    error("adding action versus self");
  if(player2 < player1)
    history[time*ROUND_SIZE + player1*MAX_POP + player2] = action;
  else
    history[time*ROUND_SIZE + player1*MAX_POP + player2] = action;  
  
  return;
}

void
add_action_prop(int * history, agent player1, agent player2, int time, int action)
{
  if(time < 0)
    error("adding action at negative time");
  
  history[time*ROUND_SIZE + player1*MAX_POP + player2] = action;

  return;
}

/*
** gets the action that one player did to another at a given time
*/

int
get_action(int * history, agent player1, agent player2, int time)
{
  int action;
  if(time < 0)
    return COOPERATE;   /* they didn't defect at least... */

  if(player1 == player2){
    return COOPERATE;     /* this will only happen in evaluation when   */
                          /* a variable is assigned in this strange way */
  }

  action = history[time*ROUND_SIZE + player1*MAX_POP + player2];
  
  if (action == NONE){
    printf("get_action: there was no interaction between %d and %d @ %d\n",
           player1, player2, time);
    /* this is indeed a bug and should be impossible */
    return COOPERATE;
  }

  return action;
}

int
get_action_prop(int * history, agent player1, agent player2, int time)
{
  int action;
  
  if(time < 0){
    return COOPERATE; /* they did not defect... */
  }

  action = history[time*ROUND_SIZE + player1*MAX_POP + player2];

  if (action == NONE){
    printf("strange NONE action found, in %d vs. %d @ t = %d\n", player1, player2, time);
    /* this should only happen if player1 == player2 and have 1 (or 0 ?) pop */
    action = COOPERATE;
  }
  
  return action;
}

void
get_Results(int * history, Result * results, int pop)
{
  int time;
  int row, column;
  int row_action, column_action;

  for(row = 0; row < pop; row++){
    results[row].t = 0;
    results[row].r = 0;
    results[row].p = 0;
    results[row].s = 0;
    results[row].score = 0;
  }

  /* for(time = 0; history[time * ROUND_SIZE] != NONE; time++)*/
  /* this is here for historical purposes. I can't see why I used it */

  for(time = 0; time < GAME_LENGTH; time++)
    for(row = 0; row < pop; row++)
      for(column = 0; column < pop; column++)
        if(row!=column){
          row_action = get_action(history,row,column,time);
          column_action = get_action(history,column,row,time);
          if(row_action==DEFECT && column_action==COOPERATE){
            results[row].t++;
            results[row].score += TEMPTATION;
          }
          else if(row_action==COOPERATE && column_action==COOPERATE){
            results[row].r++;
            results[row].score += REWARD;
          }
          else if(row_action==DEFECT && column_action==DEFECT){
            results[row].p++;
            results[row].score += PUNISHMENT;
          }
          else if(row_action==COOPERATE && column_action==DEFECT){
            results[row].s++;
            results[row].score += SUCKER;
          }
        }
  return;
}

void
get_Results_prop(int * history, Result * results, int * quantities, int pop)
{
  int time;
  int mod = 0; /* hacky */
  int row, column;
  int row_action, column_action;

  for(row = 0; row < pop; row++){
    results[row].t = 0;
    results[row].r = 0;
    results[row].p = 0;
    results[row].s = 0;
    results[row].score = 0;
  }


  /* for(time = 0; history[time * ROUND_SIZE] != NONE; time++)       */
  /* this is here for historical purposes. I can't see why I used it */

  for(time = 0; time < GAME_LENGTH; time++)
    for(row = 0; row < pop; row++)
      for(column = 0; column < pop; column++){
        if((quantities[column] >= 1 && quantities[row] >= 1 && row != column)
             || (row == column && quantities[row] >= 2)) {
          row_action = get_action_prop(history,row,column,time);
          column_action = get_action_prop(history,column,row,time);  
          if(row == column)
            mod = 1;
          else
            mod = 0;
          if(row_action==DEFECT && column_action==COOPERATE){
            results[row].t += quantities[column] - mod;
            results[row].score += (quantities[column] - mod) * TEMPTATION;
          }
          else if(row_action==COOPERATE && column_action==COOPERATE){
            results[row].r += quantities[column] - mod;
            results[row].score += (quantities[column] - mod) * REWARD;
          }
          else if(row_action==DEFECT && column_action==DEFECT){
            results[row].p += quantities[column] - mod;
            results[row].score += (quantities[column] - mod) * PUNISHMENT;
          }
          else if(row_action==COOPERATE && column_action==DEFECT){
            results[row].s += quantities[column] - mod;
            results[row].score += (quantities[column] - mod) * SUCKER;
          }
        }
      }
  return;
}

/*
** Updates the proportions of each agent type at the end of an SIPD game
** according to how well they went.
*/

void
update_proportions(Result * results, int * quantities, int pop)
{
  int player;
  int total_quantity = 0;
  int new_total_quantity = 0;
  int total_score = 0;
  double min;
  int most_deserving;
  double average_score;
  double real_new_quantity;
  double diffs[MAX_POP];

  for(player = 0; player < pop; player++){
    total_quantity += quantities[player];
    total_score += results[player].score * quantities[player];
  }

  average_score = total_score / total_quantity;

  for(player = 0; player < pop; player++){
    real_new_quantity = ((double) quantities[player]
                            * ((double) results[player].score / average_score));
    quantities[player] = (int) real_new_quantity; 
    diffs[player] = quantities[player] - real_new_quantity;
  }

  for(player = 0; player < pop; player++){
    new_total_quantity += quantities[player];
  }

  /* fill up the empty spots caused by rounding */
  while(new_total_quantity < total_quantity){
    min = 0;
    most_deserving = NONE;
    for(player = 0; player < pop; player++)
      if(diffs[player] < min){
	min = diffs[player];
	most_deserving = player;
      }
      if(most_deserving != NONE){
	quantities[most_deserving]++;
	diffs[most_deserving]++;
	new_total_quantity++;
      }
      else
	debug("problem getting new proportions");
  }

  return;
}



/***********************************************************************

      formula functions:

***********************************************************************/

Formula *
new_Formula(int type)
{
  Formula * formula;

  if(type == NONE)
    error("attempting to make formula with improper type");

  if ((formula = malloc(sizeof(Formula))) == NULL)
    error("malloc ran out of memory making formula");
  formula->type         = type;
  formula->left         = NULL;
  formula->right        = NULL;
  formula->player1      = NONE;
  formula->player2      = NONE;
  formula->time         = 0;
  formula->time_mod     = 0;
  formula->chance       = 0;
  return formula;
}

void
free_Formula(Formula * formula)
{
  if(formula != NULL){
    free_Formula(formula->left);
    free_Formula(formula->right);
    free(formula);
  }

  return;
}

void
init_random_Formulae(Formula * formulae[], int pop)
{
  char agents[MAX_AGENT_SYMBOLS];
  char times[MAX_TIME_SYMBOLS];
  int player;

  for(player = 0; player < pop; player++){
    init_variables(agents,times);
    formulae[player] = random_Formula(agents, times, 1, 1);
  }
  return;
}

Formula *
random_Formula(string agents, string times, int depth, int quantification)
{
  Formula * formula;
  int type = NONE;
  int variable, variable2;
  char new_agents[MAX_AGENT_SYMBOLS];
  char new_times[MAX_TIME_SYMBOLS];

  strcpy(new_agents,agents);
  strcpy(new_times,times);

  if(depth == MAX_DEPTH)
    variable = 0;
  else if(quantification >= MAX_QUANTIFICATION)
    variable = (rand() % (AND_CHANCE +
                          OR_CHANCE +
                          NOT_CHANCE +
                          ATOM_CHANCE));
  
  else
    variable = (rand() % (AND_CHANCE +
                          OR_CHANCE  +
                          NOT_CHANCE +
                          ALL_A_CHANCE +
                          EXISTS_A_CHANCE +
                          ALL_T_CHANCE +
                          EXISTS_T_CHANCE +
                          ATOM_CHANCE));

  if(variable >= 0 && variable < ATOM_CHANCE)
    type = ATOM;
  variable -= ATOM_CHANCE;
  if(variable >= 0 && variable < AND_CHANCE)
    type = AND;
  variable -= AND_CHANCE;
  if(variable >= 0 && variable < OR_CHANCE)
    type = OR;
  variable -= OR_CHANCE;
  if(variable >= 0 && variable < NOT_CHANCE)
    type = NOT;
  variable -= NOT_CHANCE;
  if(variable >= 0 && variable < ALL_A_CHANCE)
    type = ALL_AGENTS;
  variable -= ALL_A_CHANCE;
  if(variable >= 0 && variable < EXISTS_A_CHANCE)
    type = EXISTS_AGENT;
  variable -= EXISTS_A_CHANCE;
  if(variable >= 0 && variable < ALL_T_CHANCE)
    type = ALL_TIMES;
  variable -= ALL_T_CHANCE;
  if(variable >= 0 && variable < EXISTS_T_CHANCE)
    type = EXISTS_TIME;
  variable -= EXISTS_T_CHANCE;

  formula = new_Formula(type);  
  switch(formula->type){
    case AND:
    case OR:
      formula->right = random_Formula(new_agents, new_times,
                                      depth+1, quantification);
    case NOT:
      formula->left = random_Formula(new_agents, new_times,
                                      depth+1, quantification);
      break;
    case ALL_AGENTS:
    case EXISTS_AGENT:
      formula->variable = create_new_agent_var(new_agents);
      formula->left = random_Formula(new_agents, new_times, depth+1,
                                     quantification * POP_FOR_RANDOM_FORMULAE);
      break;
    case ALL_TIMES:
    case EXISTS_TIME:
      formula->variable = create_new_time_var(new_times);
      formula->left = random_Formula(new_agents, new_times, depth+1,
                                     quantification * GAME_LENGTH);
      break;
    case ATOM:
      formula->player1 = random_agent(new_agents,NONE);
      formula->player2 = random_agent(new_agents,formula->player1);
      formula->time = random_time(new_times);
      formula->time_mod = rand() % 3 - 2;

      break;
    default:
      error("weird type in 'random formula'");
  }

  return formula;
}

/*
** chooses a random agent from a list of allowable variables / constants
** making sure it is not the same as the forbidden agent
** (usually the other agent referred to in the atom)
*/

agent
random_agent(string agents, agent forbidden)
{
  agent new_agent = forbidden;
  int index;

  while(new_agent==forbidden){
    index = rand() % strlen(agents);
    new_agent = agents[index];
    if(new_agent=='r')
      new_agent = ROW;
    else if(new_agent=='c')
      new_agent = COLUMN;
  } 

  return new_agent;
}

/*
** chooses a random time from a list of allowable variables / constants
*/

int
random_time(string times)
{
  int new_time;
  int index;

  index = rand() % strlen(times);
  new_time = times[index];
  if(new_time=='p')
    new_time = PREV; 

  return new_time;
}

void
init_variables(string agents, string times)
{
  int i;

  agents[0] = 'r';
  agents[1] = 'c';
  for(i = 2; i < MAX_AGENT_SYMBOLS; i++)
    agents[i] = '\0';

  times[0] = 'p';
  for(i = 1; i < MAX_TIME_SYMBOLS; i++)
    times[i] = '\0';

  return;
}

void
add_variable(string variables, int variable)
{
  int length;

  length = strlen(variables);
  variables[length] = variable;
  variables[length+1] = '\0';

  return;
}


/*
** corrects all the formulae with regards to illegal variables by
** calling fix_Formula on each of them
*/

void
fix_Formulae(Formula * formulae[], int pop, int force_simplification)
{
  char agents[MAX_AGENT_SYMBOLS];
  char times[MAX_TIME_SYMBOLS];
  int player;
  

  for(player = 0; player < pop; player++){
    init_variables(agents,times);

    /*
    printf("pre-fixing <");
    print_Formula(formulae[player]);
    printf(">\n");
    */

    fix_Formula(formulae[player], agents, times);
    
    /*
    printf("post-fixing <");
    print_Formula(formulae[player]);
    printf(">\n");
    */

    if(SIMPLIFY_FORMULAE || force_simplification)
      simplify_Formula(&(formulae[player]));

    /*
    printf("post-simplifying <");
    print_Formula(formulae[player]);
    printf(">\n");
    */
  }

  return;
}

/*
** goes through a formula recursively to find variables that have not
** been quantified and replaces them with random allowable variables or
** constants
*/

void
fix_Formula(Formula * formula, string agents, string times)
{
  char new_agents[MAX_AGENT_SYMBOLS];
  char new_times[MAX_TIME_SYMBOLS];
  char new_variable;

  strcpy(new_agents,agents);
  strcpy(new_times,times);

  switch(formula->type){
    case AND:
    case OR:
      fix_Formula(formula->right, new_agents, new_times);
    case NOT:
      fix_Formula(formula->left,  new_agents, new_times);
      break;
    case ALL_AGENTS:
    case EXISTS_AGENT:
      if(strchr(new_agents, formula->variable) != NULL) {
        new_variable = create_new_agent_var(new_agents);
        replace_var(formula, formula->variable, new_variable);
        fix_Formula(formula->left, new_agents, new_times);
      }
      else {
        add_variable(new_agents, formula->variable);
        fix_Formula(formula->left, new_agents, new_times);
      }
      break;
    case ALL_TIMES:
    case EXISTS_TIME:
      if(strchr(new_times, formula->variable) != NULL) {
        new_variable = create_new_time_var(new_times);
        replace_var(formula, formula->variable, new_variable);
        fix_Formula(formula->left, new_agents, new_times);
      }
      else {
        add_variable(new_times, formula->variable);
        fix_Formula(formula->left, new_agents, new_times);
      }
      break;
    case ATOM:
      if(formula->player1 != ROW && formula->player1 != COLUMN
                        && strchr(new_agents,formula->player1)==NULL)
        formula->player1 = random_agent(new_agents,formula->player2);
      if(formula->player2 != ROW && formula->player2 != COLUMN
                        && strchr(new_agents,formula->player2)==NULL)
        formula->player2 = random_agent(new_agents,formula->player1);
      if(formula->time != PREV && strchr(new_times,formula->time)==NULL)
        formula->time = random_time(new_times);
      break;
    default:
      error("weird type in 'fix formula'");
  }

  return;
}

void
replace_var(Formula * formula, int old_var, int new_var)
{
  switch(formula->type){
    case AND:
    case OR:
      replace_var(formula->right, old_var, new_var);
    case NOT:
      replace_var(formula->left, old_var, new_var);
      break;
    case ALL_AGENTS:
    case EXISTS_AGENT:
    case ALL_TIMES:
    case EXISTS_TIME:
      if(formula->variable == old_var)
        formula->variable = new_var;
      replace_var(formula->left, old_var, new_var);
      break;
    case ATOM:
      if(formula->player1 == old_var)
        formula->player1 = new_var;
      if(formula->player2 == old_var)
        formula->player2 = new_var;
      if(formula->time == old_var)
        formula->time = new_var;
      break;
    default:
      error("weird type in 'replace var'");
  }
  return;
}

int
useless_var(Formula * formula, int var)
{
  int useless = TRUE;

  switch(formula->type){
    case AND:
    case OR:
      useless = (useless_var(formula->right, var) &&
                 useless_var(formula->left, var));
      break;
    case NOT:
    case ALL_AGENTS:
    case EXISTS_AGENT:
    case ALL_TIMES:
    case EXISTS_TIME:
      useless = useless_var(formula->left, var);
      break;
    case ATOM:
      if(formula->player1 == var)
        useless = FALSE;
      if(formula->player2 == var)
        useless = FALSE;
      if(formula->time == var)
        useless = FALSE;
      break;
    default:
      error("weird type in 'finding useless variables'");
  }

  return useless;
}


void
simplify_Formula(Formula ** formula_ptr)
{
  Formula * fake_parent_formula = new_Formula(NOT);
  
  simplify_Formula_2(fake_parent_formula,*formula_ptr,LEFT);
  if(fake_parent_formula->left != NULL)
    *formula_ptr = fake_parent_formula->left;

  fake_parent_formula->left = NULL;
  free_Formula(fake_parent_formula);    
  
  return;
}

void
simplify_Formula_2(Formula * parent_formula, Formula * formula, int direction)
{
  Formula * temp_formula;

  switch(formula->type){
    case AND:
    case OR:
      simplify_Formula_2(formula, formula->right, RIGHT);
      simplify_Formula_2(formula, formula->left, LEFT);
      break;
    case NOT:
      if(formula->left->type == NOT){          /* double neg elim */
        temp_formula = formula->left->left;
        formula->left->left = NULL;
        free_Formula(formula); 
        if(direction == LEFT) {
          parent_formula->left = temp_formula;
          simplify_Formula_2(parent_formula, temp_formula, LEFT);
        }
	else {
          parent_formula->right = temp_formula;
          simplify_Formula_2(parent_formula, temp_formula, RIGHT);
        }
      }
      else
        simplify_Formula_2(formula, formula->left, LEFT);
      break;
    case ALL_AGENTS:
    case EXISTS_AGENT:
    case ALL_TIMES:
    case EXISTS_TIME:
      if(useless_var(formula->left, formula->variable)){  /* excess quantifier elim */
        temp_formula = formula->left;
        formula->left = NULL;
        free_Formula(formula); 
        if(direction == LEFT){
          parent_formula->left = temp_formula;
          simplify_Formula_2(parent_formula, temp_formula, LEFT);
        }
        else {
          parent_formula->right = temp_formula;
          simplify_Formula_2(parent_formula, temp_formula, RIGHT);
        }
      }
      else
        simplify_Formula_2(formula, formula->left, LEFT);
      break;
    case ATOM:
      break;
    default:
      error("weird type in 'simplify formula'");
  }
  return;
}



/*
** Creates a new agent variable from those as yet undefined
** it adds this value to the list of defined agents and also returns it
*/

int
create_new_agent_var(string agents)
{
  int variable;

  if(strlen(agents) == MAX_AGENT_SYMBOLS)
    error("exhausted available agent variables");

  variable = rand() % MAX_AGENT_VARS + 'd';
  while(strchr(agents, variable) != NULL)
    variable = rand() % MAX_AGENT_VARS + 'd';
  add_variable(agents, variable);

  return variable;
}

/*
** Creates a new time variable from those as yet undefined
** it adds this value to the list of defined agents and also returns it
*/

int
create_new_time_var(string times)
{
  int variable;

  if(strlen(times) == MAX_TIME_SYMBOLS)
    error("exhausted available time variables");

  variable = rand() % MAX_TIME_VARS + 's';
  while(strchr(times, variable) != NULL)
    variable = rand() % MAX_TIME_VARS + 's';
  add_variable(times, variable);

  return variable;
}

/*****************************************************************

      genetic functions:

*****************************************************************/

void
update_Formulae(Result * results, Formula ** formulae, int pop)
{
  int player;
  int parent1;
  int parent2;
  Formula * new_formulae[MAX_POP];

  printf("updating formulae\n");

  for(player = 0; player < pop; player++){
    parent1 = choose_parent(results, pop);
    parent2 = choose_parent(results, pop);
    new_formulae[player] = cross_over(formulae[parent1],formulae[parent2]);
 
    if(rand() % INVERSE_MUT_RATE == 0)
      mutate_Formula(new_formulae[player]);
    /*
    print_Formula(formulae[parent1]);
    printf("\n");
    print_Formula(formulae[parent2]);
    printf("\n");
    print_Formula(new_formulae[player]);
    printf("\n\n");
    */
  }

  for(player = 0; player < pop; player++){
    free_Formula(formulae[player]);
    formulae[player] = new_formulae[player];
  }

  fix_Formulae(formulae, pop, FALSE);

  return;
}



Formula *
cross_over(Formula * parent1, Formula * parent2)
{
  Formula * child = copy_Formula(parent1);
  Formula * point1;
  Formula * point2;
  Formula * point2_next;

  /* choose a point in parent2 to take a subtree */

  point2 = choose_random_Formula(parent2);

  if(point2 == NULL)
    point2_next = parent2;
  else if(point2->type <= OR && (rand() % 2))
    point2_next = point2->right;
  else
    point2_next = point2->left;

  /* choose a point in the child tree to replace with the subtree from parent2 */

  point1 = choose_random_Formula(child);

  if(point1 == NULL){
    free_Formula(child);
    child = copy_Formula(point2_next);
  }
  else if(point1->type <= OR && (rand() % 2)){
    free_Formula(point1->right);
    point1->right = copy_Formula(point2_next);
  }
  else{
    free_Formula(point1->left);
    point1->left = copy_Formula(point2_next);
  }

  return child;
}

/*
** the chance of choosing a given agent is its score divided by the
** total score.
*/

agent
choose_parent(Result * results, int pop)
{
  int player;
  int parent = NONE;
  int total_score = 0;
  int running_score = 0;
  int random_num;
  int found;

  for(player = 0; player < pop; player++)
    total_score += results[player].score;

  random_num = rand() % total_score;

  found = FALSE;
  for(player = 0; player < pop && !found; player++){
    running_score += results[player].score;
    if(random_num < running_score){
      parent = player;
      found = TRUE;
    }
  }

  if(parent==NONE)
    error("failure in choosing parent");

  return parent;
}

Formula *
copy_Formula(Formula * formula)
{
  Formula * new_formula;

  if(formula == NULL)
    new_formula = NULL;
  else{
    new_formula = new_Formula(formula->type);
    new_formula->variable = formula->variable;
    new_formula->player1  = formula->player1;
    new_formula->player2  = formula->player2;
    new_formula->time     = formula->time;
    new_formula->time_mod = formula->time_mod;
    new_formula->chance   = formula->chance;
    new_formula->left     = copy_Formula(formula->left);
    new_formula->right    = copy_Formula(formula->right);
  }
  
  return new_formula;
}


/*
** chooses a random node from a formula.
** the node that will be used in crossover is a child of this node,
** therefore this node will not be a leaf
** and to allow the root to be chosen, this function can return NULL
** which indicates that this is to be the case
*/


Formula *
choose_random_Formula(Formula * formula)
{
  Formula_List * formula_list;
  int random_number;
  int i;
  int length;

  formula_list = flatten_Formula(formula,NULL);

  formula_list = cons_Formula(NULL,formula_list);
  length = length_Formula_List(formula_list);

  random_number = rand() % length;

  for(i = 0; i < random_number; i++)
    formula_list = formula_list->next;

  return formula_list->formula;
}

int
length_Formula_List(Formula_List * formula_list)
{
  int length;
  Formula_List * curr = formula_list;

  for(length = 0; curr != NULL; length++)
    curr = curr->next;

  return length;
}

Formula_List *
flatten_Formula(Formula * formula, Formula_List * formula_list)
{
  Formula_List * new_formula_list = NULL;

  if(formula->type <= OR){               /* nodes with two branches */
    new_formula_list = cons_Formula(formula, formula_list);
    new_formula_list = flatten_Formula(formula->left, new_formula_list);
    new_formula_list = flatten_Formula(formula->right, new_formula_list);
  }
  else if(formula->type <= EXISTS_TIME){ /* nodes with one branch */
    new_formula_list = cons_Formula(formula, formula_list);
    new_formula_list = flatten_Formula(formula->left, new_formula_list);
  }
  else
    new_formula_list = formula_list;

  return new_formula_list;
}

Formula_List *
cons_Formula(Formula * formula, Formula_List * formula_list)
{
  Formula_List * new_head;
  if ((new_head = malloc(sizeof(Formula_List))) == NULL)
    error("malloc ran out of memory performing crossover");

  new_head->formula = formula;
  new_head->next = formula_list;

  return new_head;
}



/* not called yet */

void
mutate_Formula(Formula * formula)
{
  char agents[MAX_AGENT_SYMBOLS];
  char times[MAX_TIME_SYMBOLS];
  
  init_variables(agents, times);

  mutate_an_atom(formula, agents, times);

  return;
}


/* note that it biases in favor of less nested ands/ors - change this? */

void
mutate_an_atom(Formula * formula, string agents, string times)
{
  char new_agents[MAX_AGENT_SYMBOLS];
  char new_times[MAX_TIME_SYMBOLS];

  strcpy(new_agents,agents);
  strcpy(new_times,times);

  switch(formula->type){
    case AND:
    case OR:
      if(rand() % 2 == 1)
        mutate_an_atom(formula->left, new_agents, new_times);
      else
        mutate_an_atom(formula->right, new_agents, new_times);
      break;
    case NOT:
      mutate_an_atom(formula->left, new_agents, new_times);
      break;
    case ALL_AGENTS:
    case EXISTS_AGENT:
      add_variable(new_agents, formula->variable);
      mutate_an_atom(formula->left, new_agents, new_times);
      break;
    case ALL_TIMES:
    case EXISTS_TIME:
      add_variable(new_times, formula->variable);
      mutate_an_atom(formula->left, new_agents, new_times);
      break;
    case ATOM:
      mutate_atom(formula, new_agents, new_times);
      break;
    default:
      error("Invalid contructor in formula (found during mutation)");
  }
  

  return;
}

void
mutate_atom(Formula * atom, string agents, string times)
{
  int substitution;
  int location;

     location = rand() % 4 + 1;
      if(location == 1)
        atom->player1 = random_agent(agents, atom->player2);
      else if(location == 2)
        atom->player2 = random_agent(agents, atom->player1);
      else if(location == 3)
        atom->time = random_time(times);
      else
        atom->time_mod = - ((rand() % 3)); 
      return;
}




/**************************************************************************

      input functions:

**************************************************************************/

/*
** gets the formulae from the files listed in standard input
*/

int
init_Formulae(Formula * formulae[])
{
  char buffer[MAX_FILE_LENGTH];
  int pop = 0;
  printf("getting players...\n");
  while(scanf("%s",buffer) != -1){
    formulae[pop] = read_Formula(buffer);
    pop++;
  }
  return pop;
}

/*
** gets the formulae / quantity pairs from standard input
*/

int
init_Formulae_prop(Formula * formulae[], int quantities[])
{
  char buffer[MAX_FILE_LENGTH];
  int pop = 0;
  int i;

  printf("getting players...\n");
  while(scanf("%s",buffer) != -1){         /* scan the strategy name */
    formulae[pop] = read_Formula(buffer);
    scanf("%s", buffer);                   /* scan the 'x'           */
    scanf("%d", &quantities[pop]);         /* scan the quantity      */
    pop++;
  }
  return pop;
}


/*
** reads in a formula from the named file
*/

Formula *
read_Formula(const char * file_name)
{
  Formula * formula;
  FILE * file;
  char path_name[MAX_FILE_NAME_LENGTH * 2] = {0};
  char buffer[MAX_FILE_LENGTH];
  char c;
  int i = 0;

  strcat(path_name, STRATEGY_LOCATION);
  strcat(path_name, file_name);
  file = fopen(path_name,"r"); 

  while((c = fgetc(file)) != EOF)
    if(c!=' ' && c!='\t' && c!='\n'){
      if(c == '%'){
        while(c != '\n' && c != EOF)
          c = fgetc(file);
        if(c == EOF)
          break;
      }
      else{
        buffer[i] = c;
        putchar(buffer[i]);
        i++;
      }
    }

  putchar('\n');
  formula = make_Formula(buffer,0,i-1);

  return formula;
}




/*
** recursively turns a string into a formula tree
*/

Formula *
make_Formula(char * buffer, int low, int high)
{
  Formula * formula = NULL; /* gnuc */
  int location = low;
 
  location = find_conjunct(buffer, low, high);

  if(location == NONE)
    switch(buffer[low]){
    case '(':
      return make_Formula(buffer, low+1, high-1);
    case 'D':
      formula = new_Formula(ATOM);
      if(buffer[low+1] == 'r')
        formula->player1 = ROW;
      else if(buffer[low+1] == 'c')
        formula->player1 = COLUMN;
      else
        formula->player1 = (int) buffer[low+1];
      if(buffer[low+2] == 'r')
        formula->player2 = ROW;
      else if(buffer[low+2] == 'c')
        formula->player2 = COLUMN;
      else
        formula->player2 = (int) buffer[low+2];
      if(buffer[low+3] == 'p')
        formula->time = PREV;
      else if(buffer[low+3]>='a' && buffer[low+3]<='z')
        formula->time = (int) buffer[low+3];
      formula->time_mod = make_time_mod(buffer, low, high);
      return formula;    
    case 'R':
      formula = new_Formula(RANDOM);
      formula->chance = get_chance(buffer, low, high);
      return formula;
    case 'A':
      if(buffer[low+1]!='\''){
        formula = new_Formula(ALL_AGENTS);
        formula->variable = (int) buffer[low+1];
        formula->left = make_Formula(buffer, low+2, high);
      }
      else{
        formula = new_Formula(ALL_TIMES);
        formula->variable = (int) buffer[low+2];
        formula->left = make_Formula(buffer, low+3, high);
      }
      break;
    case 'E':
      if(buffer[low+1]!='\''){
        formula = new_Formula(EXISTS_AGENT);
        formula->variable = (int) buffer[low+1];
        formula->left = make_Formula(buffer, low+2, high);
      }
      else{
        formula = new_Formula(EXISTS_TIME);
        formula->variable = (int) buffer[low+2];
        formula->left = make_Formula(buffer, low+3, high);
      }
      break;
    default:
      fprintf(stderr,"low = %d, high = %d\n",low, high);
      error("no atoms or operands (?) in the current scope");
      break;
    }
  else /* if an operand (?) was found */
    switch(buffer[location]){
      case '&':
        formula = new_Formula(AND);
        formula->left  = make_Formula(buffer, low, location-1);
        formula->right = make_Formula(buffer, location+1, high);
        break;
      case '|':
        formula = new_Formula(OR);
        formula->left  = make_Formula(buffer, low, location-1);
        formula->right = make_Formula(buffer, location+1, high);
        break;
      case '~':
        formula = new_Formula(NOT);
        formula->left = make_Formula(buffer, low+1, high);
        break;
      default:
        fprintf(stderr,"<<%c>>",buffer[location]); 
        error("unexpected character while reading formula");
    }
  return formula;
}

int
make_time_mod(char * buffer, int low, int high)
{
  char num_buffer[MAX_INT_LENGTH];
  int i;

  if(low+3==high && buffer[low+3]=='p')
    return 0;
  if(low+3==high && buffer[low+3]!='p')
    low--;

  for(i=0;i<MAX_INT_LENGTH;i++)
    num_buffer[i] = 0;      
  for(i=low+4; i<=high; i++)
    num_buffer[i-low-4] = buffer[i];

  return atoi(num_buffer);
}

double
get_chance(char * buffer, int low, int high)
{
  char num_buffer[MAX_INT_LENGTH];
  int i;

  for(i=0;i<MAX_INT_LENGTH;i++)
    num_buffer[i] = 0;      
  for(i=low+1; i<=high; i++)
    num_buffer[i-low-1] = buffer[i];

  return atof(num_buffer);
}


/*
** goes through the text looking for operators in the current scope.
** it needs to find them in order of weakest binding to strongest
** Therefore, it looks for OR from back to front, then AND, then NOT
*/

int
find_conjunct(char * buffer, int low, int high)
{
  int loc;
  int found = FALSE;
  int count = 0;

  for(loc = high; !found && loc>=low; loc--){
    if(buffer[loc]=='(')
      count++;
    else if(buffer[loc]==')')
      count--;
    if(buffer[loc]=='|' && count==0)
      found = TRUE;
  }

  if(!found)
    for(loc = high; !found && loc>=low; loc--){
    if(buffer[loc]=='(')
      count++;
    else if(buffer[loc]==')')
      count--;
    if(buffer[loc]=='&' && count==0)
      found = TRUE;
    }

  if(!found)
    for(loc = high; !found && loc>=low; loc--){
    if(buffer[loc]=='(')
      count++;
    else if(buffer[loc]==')')
      count--;
    if(buffer[loc]=='~' && count==0)
      found = TRUE;
    }

  loc++; /* the silly FOR loop goes one too far */

  if(!found)
    return NONE;
  else
    return loc;
}



/********************************************************************

      printing functions:

********************************************************************/


void
print_Formula(Formula * formula)
{
  if(formula == NULL)
    printf("NULL");
  else
    switch(formula->type){
      case AND:
        printf("(");
        print_Formula(formula->left);
        printf("&");
        print_Formula(formula->right);
        printf(")");
        break;
      case OR:
        printf("(");
        print_Formula(formula->left);
        printf("|");
        print_Formula(formula->right);
        printf(")");
        break;
      case NOT:
        printf("(~");
        print_Formula(formula->left);
        printf(")");
        break;
      case ALL_AGENTS:
        printf("(A%c",(char) formula->variable);
        print_Formula(formula->left);
        printf(")");
        break;
      case EXISTS_AGENT:
        printf("(E%c",(char) formula->variable);
        print_Formula(formula->left);
        printf(")");
        break;
      case ALL_TIMES:
        printf("(A'%c",(char) formula->variable);
        print_Formula(formula->left);
        printf(")");
        break;
      case EXISTS_TIME:
        printf("(E'%c",(char) formula->variable);
        print_Formula(formula->left);
        printf(")");
        break;
      case ATOM:
        printf("D");
        if(formula->player1 == ROW)
          printf("r");
        else if(formula->player1 == COLUMN)
          printf("c");
        else
          printf("%c",(char) formula->player1);
        if(formula->player2 == ROW)
          printf("r");
        else if(formula->player2 == COLUMN)
          printf("c");
        else
          printf("%c",(char) formula->player2);
        if(formula->time == PREV){
          printf("p");
          if(formula->time_mod != 0)
            printf("%d",formula->time_mod);
        }
        else if(formula->time != 0){
          printf("%c",(char) formula->time);
          if(formula->time_mod != 0)
            printf("%d",formula->time_mod);
        }
        else
          printf("%d",formula->time_mod);
        break;
      case RANDOM:
        printf("R%f",formula->chance);
        break;
      default:
        fprintf(stderr,"type = %d\n",formula->type);
        fprintf(stderr,"player1 = %d\n",formula->player1);
        fprintf(stderr,"player2 = %d\n",formula->player2);
        fprintf(stderr,"time = %d\n",formula->time);
        error("Invalid contructor in formula (while printing)");
    }
  return;
}

void
print_Formulae(Formula ** formulae, int pop)
{
  int player;
  for(player = 0; player < pop; player++){
    printf("%d -------------------------------------------\n",player);
    print_Formula(formulae[player]);
    printf("\n\n");
  }
  return;
}

void
print_quantities(int * quantities, int pop)
{
  int player;
  for(player = 0; player < pop; player++){
    printf("%d -------------------------------------------\n",player);
    printf("%d", quantities[player]);
   
    printf("\n\n");
  }
  return;

}

void
print_actions(int * history, int pop, agent row)
{
  agent column;
  int time;
  int row_action;

  printf("\n---------------------------------------------------\n");
  printf("actions for player %d\n",row);
  printf("time\t");
  for(column = 0; column < pop; column++)
    if(row!=column)
      printf("%d\t",column);
  printf("\n---------------------------------------------------\n");

  for(time = 0; history[time * ROUND_SIZE] != NONE; time++) {     
    printf("%d\t",time);
    for(column = 0; column < pop; column++)
      if(row!=column){
        row_action = get_action(history,row,column,time);
        printf("%d",row_action);
        row_action = get_action(history,column,row,time);
        printf("-%d\t",row_action);
      }
    printf("\n");
  }
  return;
}

void
print_Results(Result * results, int pop)
{
  agent row;

  printf("---------------------------------------------------\n");
  printf("  player       t       r       p       s   score\n");
  printf("---------------------------------------------------\n");
  for(row = 0; row < pop; row++)
    printf("%8d%8d%8d%8d%8d%8d\n",row, results[row].t,
         results[row].r, results[row].p, results[row].s, results[row].score);
  printf("---------------------------------------------------\n");
  return;
}

void
print_Results_prop(Result * results, int * quantities, int pop)
{
  agent row;

  printf("-------------------------------------------------------------\n");
  printf("  player       t       r       p       s   score   quantity\n");
  printf("-------------------------------------------------------------\n");
  for(row = 0; row < pop; row++)
    printf("%8d%8d%8d%8d%8d%8d%8d\n",row, results[row].t,
         results[row].r, results[row].p, results[row].s, results[row].score, quantities[row]);
  printf("-------------------------------------------------------------\n");
  return;
}

/* this is really a debugging function, printing a temporary data structure */

void
print_Formula_List(Formula_List * formula_list)
{
  Formula_List * curr = formula_list;

  printf("\n[ ");

  while(curr != NULL){
    printf(" , ");
    print_Formula(curr->formula);
    curr = curr->next;
  }

  printf(" ]\n");

  return;
}

/***********************************************************************

  utility functions:

***********************************************************************/

/*
** prints a string to stderr and exits
*/

void
error(const char * message)
{
  fprintf(stderr,"Error: %s\n",message);
  exit(EXIT_FAILURE);
  return;
}

/*
** prints a string to stderr if DEBUG = TRUE
*/

void
debug(const char * message)
{
  if(DEBUG)
    fprintf(stderr,"%s\n",message);
  return;
}




